import time
import gspread
import pandas as pd
from gspread import utils
from oauth2client.service_account import ServiceAccountCredentials
from tenacity import (
    retry,
    wait_incrementing,
    stop_after_attempt,
)
from gspread_formatting import (
    set_frozen,
    format_cell_range,
    format_cell_ranges,
    CellFormat,
    Color,
    TextFormat,
    NumberFormat,
)


class GoogleDocs:
    def __init__(self, keyfile=None):
        """
        Initializes the GoogleDocs class with credentials for Google API.
        :param keyfile: Path to the JSON file containing Google service account credentials.
        Defaults to "./credentials.json" if not provided.
        """
        if keyfile is None:
            keyfile = "./credentials.json"

        scope = ["https://spreadsheets.google.com/feeds", "https://www.googleapis.com/auth/drive"]
        credentials = ServiceAccountCredentials.from_json_keyfile_name(keyfile, scope)
        self.spreadsheet = None
        self.gc = gspread.authorize(credentials)

    @staticmethod
    @retry(wait=wait_incrementing(start=60, increment=30, max=300), stop=stop_after_attempt(5))
    def safe_execute(func, *args, **kwargs):
        time.sleep(2)
        return func(*args, **kwargs)

    def format_worksheet(self, worksheet, headers):
        header_range = utils.rowcol_to_a1(1, 1) + ":" + utils.rowcol_to_a1(1, len(headers))
        header_format = CellFormat(
            backgroundColor=Color(201/255, 218/255, 248/255),
            textFormat=TextFormat(bold=True),
            horizontalAlignment="CENTER",
            wrapStrategy="WRAP"
        )
        self.safe_execute(format_cell_range, worksheet, header_range, header_format)
        worksheet.freeze(rows=1)

    def get_dataframe(self, spreadsheet_key, worksheet_index=0, worksheet_title=None) -> pd.DataFrame:
        """
        Returns a worksheet from a Google Spreadsheet as a pandas DataFrame.
        :param spreadsheet_key: ID of the Google Spreadsheet.
        :param worksheet_index: Index of the worksheet (default: 0).
        :param worksheet_title: Title of the worksheet (overrides index if set).
        :return: pandas DataFrame with the worksheet data.
        """
        spreadsheet = self.gc.open_by_key(spreadsheet_key)
        if worksheet_title:
            worksheet = spreadsheet.worksheet(worksheet_title)
        else:
            worksheet = spreadsheet.get_worksheet(worksheet_index)
        data = worksheet.get_all_values()
        return pd.DataFrame(data[1:], columns=data[0])

    def write_dataframe(
        self, dataframe, worksheet_title, default_dataframe_formatting: bool = False, spreadsheet_key=None
    ):

        if spreadsheet_key:
            self.spreadsheet = self.gc.open_by_key(spreadsheet_key)
        else:
            self.spreadsheet = self.gc.create(title="new", folder_id="")

        try:
            worksheet = self.spreadsheet.worksheet(worksheet_title)
        except gspread.WorksheetNotFound:
            worksheet = self.spreadsheet.add_worksheet(title=worksheet_title, rows=100, cols=20)

        data = [dataframe.columns.values.tolist()] + dataframe.values.tolist()

        worksheet.clear()
        worksheet.update(data)

        if default_dataframe_formatting:
            columns = dataframe.columns.values.tolist()
            self.format_worksheet(worksheet=worksheet, headers=columns)


